@@ -1,11 +1,17 @@
-use crate::common::{check_installed, create_virtualenv, maybe_mock_cargo, test_python_path};
+use crate::common::{
+    check_installed, create_named_virtualenv, create_virtualenv, maybe_mock_cargo, test_python_path,
+};
 use anyhow::{bail, Context, Result};
 #[cfg(feature = "zig")]
 use cargo_zigbuild::Zig;
 use clap::Parser;
-use maturin::{BuildOptions, PythonInterpreter};
+use fs4::fs_err3::FileExt;
+use fs_err::File;
+use maturin::{BuildOptions, PlatformTag, PythonInterpreter, Target};
+use normpath::PathExt;
+use std::collections::HashSet;
 use std::env;
-use std::path::Path;
+use std::path::{Path, PathBuf};
 use std::process::Command;
 use std::str;
 
@@ -32,25 +38,25 @@ pub fn test_integration(
     let shed = format!("test-crates/wheels/{unique_name}");
     let target_dir = format!("test-crates/targets/{unique_name}");
     let python_interp = test_python_path();
-    let mut cli = vec![
-        "build",
-        "--quiet",
-        "--manifest-path",
-        &package_string,
-        "--target-dir",
-        &target_dir,
-        "--out",
-        &shed,
+    let mut cli: Vec<std::ffi::OsString> = vec![
+        "build".into(),
+        "--quiet".into(),
+        "--manifest-path".into(),
+        package_string.into(),
+        "--target-dir".into(),
+        target_dir.into(),
+        "--out".into(),
+        shed.into(),
     ];
 
     if let Some(ref bindings) = bindings {
-        cli.push("--bindings");
-        cli.push(bindings);
+        cli.push("--bindings".into());
+        cli.push(bindings.into());
     }
 
     if let Some(target) = target {
-        cli.push("--target");
-        cli.push(target)
+        cli.push("--target".into());
+        cli.push(target.into())
     }
 
     #[cfg(feature = "zig")]
@@ -59,21 +65,77 @@ pub fn test_integration(
     let zig_found = false;
 
     let test_zig = if zig && (env::var("GITHUB_ACTIONS").is_ok() || zig_found) {
-        cli.push("--zig");
+        cli.push("--zig".into());
         true
     } else {
-        cli.push("--compatibility");
-        cli.push("linux");
+        cli.push("--compatibility".into());
+        cli.push("linux".into());
         false
     };
 
+    // One scope up to extend the lifetime
+    let venvs_dir = Path::new("test-crates")
+        .normalize()?
+        .into_path_buf()
+        .join("venvs");
+    fs_err::create_dir_all(&venvs_dir)?;
+    let cffi_provider = "cffi-provider";
+    let cffi_venv = venvs_dir.join(cffi_provider);
+
     if let Some(interp) = python_interp.as_ref() {
-        cli.push("--interpreter");
-        cli.push(interp);
+        cli.push("--interpreter".into());
+        cli.push(interp.into());
+    } else {
+        // Install cffi in a separate environment
+
+        // All tests try to use this venv at the same time, so we need to make sure only one
+        // modifies it at a time and that during that time, no other test reads it.
+        let file = File::create(venvs_dir.join("cffi-provider.lock"))?;
+        file.lock_exclusive()?;
+        let python = if !cffi_venv.is_dir() {
+            create_named_virtualenv(cffi_provider, python_interp.clone().map(PathBuf::from))?;
+            let target_triple = Target::from_target_triple(None)?;
+            let python = target_triple.get_venv_python(&cffi_venv);
+            assert!(python.is_file(), "cffi venv not created correctly");
+            let pip_install_cffi = [
+                "-m",
+                "pip",
+                "--disable-pip-version-check",
+                "--no-cache-dir",
+                "install",
+                "cffi",
+            ];
+            let output = Command::new(&python)
+                .args(pip_install_cffi)
+                .output()
+                .with_context(|| format!("pip install cffi failed with {python:?}"))?;
+            if !output.status.success() {
+                let stderr = String::from_utf8_lossy(&output.stderr);
+                let stdout = String::from_utf8_lossy(&output.stdout);
+                bail!(
+                    "Installing cffi into {} failed.\nstdout: {}\nstderr: {}",
+                    cffi_venv.display(),
+                    stdout,
+                    stderr
+                );
+            }
+            python
+        } else {
+            let target_triple = Target::from_target_triple(None)?;
+            target_triple.get_venv_python(&cffi_venv)
+        };
+        file.unlock()?;
+        cli.push("--interpreter".into());
+        cli.push(python.as_os_str().to_owned());
     }
 
     let options: BuildOptions = BuildOptions::try_parse_from(cli)?;
-    let build_context = options.into_build_context(false, cfg!(feature = "faster-tests"), false)?;
+    let build_context = options
+        .into_build_context()
+        .release(false)
+        .strip(cfg!(feature = "faster-tests"))
+        .editable(false)
+        .build()?;
     let wheels = build_context.build_wheels()?;
 
     // For abi3 on unix, we didn't use a python interpreter, but we need one here
@@ -93,14 +155,20 @@ pub fn test_integration(
     // We can do this since we know that wheels are built and returned in the
     // order they are in the build context
     for ((filename, supported_version), python_interpreter) in wheels.iter().zip(interpreter) {
-        if test_zig && build_context.target.is_linux() && !build_context.target.is_musl_target() {
+        check_for_duplicates(filename)?;
+        if test_zig
+            && build_context.target.is_linux()
+            && !build_context.target.is_musl_libc()
+            && build_context.target.get_minimum_manylinux_tag() != PlatformTag::Linux
+        {
             let rustc_ver = rustc_version::version()?;
+            let python_arch = build_context.target.get_python_arch();
             let file_suffix = if rustc_ver >= semver::Version::new(1, 64, 0) {
-                "manylinux_2_17_x86_64.manylinux2014_x86_64.whl"
+                format!("manylinux_2_17_{python_arch}.manylinux2014_{python_arch}.whl")
             } else {
-                "manylinux_2_12_x86_64.manylinux2010_x86_64.whl"
+                format!("manylinux_2_12_{python_arch}.manylinux2010_{python_arch}.whl")
             };
-            assert!(filename.to_string_lossy().ends_with(file_suffix))
+            assert!(filename.to_string_lossy().ends_with(&file_suffix))
         }
         let mut venv_name = if supported_version == "py3" {
             format!("{unique_name}-py3")
@@ -166,31 +234,36 @@ pub fn test_integration_conda(package: impl AsRef<Path>, bindings: Option<String
     // Create environments to build against, prepended with "A" to ensure that integration
     // tests are executed with these environments
     let mut interpreters = Vec::new();
-    for minor in 7..=10 {
+    for minor in 9..=12 {
         let (_, venv_python) = create_conda_env(&format!("A-maturin-env-3{minor}"), 3, minor)?;
         interpreters.push(venv_python);
     }
 
     // The first argument is ignored by clap
-    let mut cli = vec![
-        "build",
-        "--manifest-path",
-        &package_string,
-        "--quiet",
-        "--interpreter",
+    let mut cli: Vec<std::ffi::OsString> = vec![
+        "build".into(),
+        "--manifest-path".into(),
+        package_string.into(),
+        "--quiet".into(),
+        "--interpreter".into(),
     ];
     for interp in &interpreters {
-        cli.push(interp.to_str().unwrap());
+        cli.push(interp.to_str().unwrap().into());
     }
 
     if let Some(ref bindings) = bindings {
-        cli.push("--bindings");
-        cli.push(bindings);
+        cli.push("--bindings".into());
+        cli.push(bindings.into());
     }
 
     let options = BuildOptions::try_parse_from(cli)?;
 
-    let build_context = options.into_build_context(false, cfg!(feature = "faster-tests"), false)?;
+    let build_context = options
+        .into_build_context()
+        .release(false)
+        .strip(cfg!(feature = "faster-tests"))
+        .editable(false)
+        .build()?;
     let wheels = build_context.build_wheels()?;
 
     let mut conda_wheels: Vec<(PathBuf, PathBuf)> = vec![];
@@ -226,3 +299,17 @@ pub fn test_integration_conda(package: impl AsRef<Path>, bindings: Option<String
 
     Ok(())
 }
+
+/// See <https://github.com/PyO3/maturin/issues/2106> and
+/// <https://github.com/PyO3/maturin/issues/2066>.
+fn check_for_duplicates(wheel: &Path) -> Result<()> {
+    let mut seen = HashSet::new();
+    let mut reader = File::open(wheel)?;
+    // We have to use this API since `ZipArchive` deduplicates names.
+    while let Some(file) = zip::read::read_zipfile_from_stream(&mut reader)? {
+        if !seen.insert(file.name().to_string()) {
+            bail!("Duplicate file: {}", file.name());
+        }
+    }
+    Ok(())
+}
